//
// directory_reader.hpp
//
// Bill Seymour, 2024-03-08
//
// This file is in the public domain.
//
// This is the header for a class that cycles through some directory.
// It's intended to be portable to POSIX systems and Microsoft Windows.
//
// It doesn't take any action on any directory and doesn't automatically
// recurse through subdirectories, but just returns information about
// the directory entries.  Users can do whatever they want with the
// information including explicitly recursing through subdirectories.
//
// It's intended to be a singleton, so it's neither copyable nor moveable,
// and you can't derive another class from it.
//

#ifndef DIRECTORY_READER_HPP_INCLUDED
#define DIRECTORY_READER_HPP_INCLUDED

#include <string>
#include <vector>
#include <climits> // CHAR_BIT
#include <ctime>   // time_t

#ifdef _MSC_VER
  #include <io.h>
  #define DIRECTORY_READER_READ_NOEXCEPT noexcept
#else
  #include <sys/types.h>
  #include <sys/stat.h>
  #include <dirent.h>
  #define DIRECTORY_READER_READ_NOEXCEPT
#endif

static_assert(sizeof(std::time_t) * CHAR_BIT >= 64, "64-bit time_t required");

namespace dirrdr {

class directory_reader final
{
    struct info_type final
    {
        std::string prev_dir, this_dir;

      #ifdef _MSC_VER
        intptr_t handle;
        typedef _finddatai64_t buftype;
      #else
        DIR* handle;
        struct dirent* dirp;
        typedef struct stat buftype;
        bool do_read();
      #endif

        buftype databuf;

        info_type();
        ~info_type();

        // move-only
        info_type(const info_type&) = delete;
        info_type& operator=(const info_type&) = delete;

        info_type(info_type&&) noexcept;
        info_type& operator=(info_type&&) noexcept;
    };

    std::vector<info_type> data;

public:
    directory_reader() : data() { }
    ~directory_reader()
    {
        while (!data.empty())
        {
            close(); // does a pop_back()
        }
    }

    // neither copyable nor moveable
    directory_reader(const directory_reader&) = delete;
    directory_reader(directory_reader&&) = delete;
    directory_reader& operator=(const directory_reader&) = delete;
    directory_reader& operator=(directory_reader&&) = delete;

    directory_reader& open(const char* = nullptr);
    directory_reader& open(const std::string&);

    bool rewind() DIRECTORY_READER_READ_NOEXCEPT;
    bool read_next() DIRECTORY_READER_READ_NOEXCEPT;

    bool at_end() const noexcept { return data.empty(); }
    auto dir_count() const noexcept { return data.size(); }

    void close();

    const std::string& directory() const noexcept;

    const char* entry_name() const noexcept;

    unsigned long long file_size() const noexcept;

    bool is_regular_file() const noexcept;
    bool is_directory() const noexcept;

    typedef info_type::buftype buftype;
    const buftype& more_data() const noexcept;
};

} // namespace dirrdr

#endif // DIRECTORY_READER_HPP_INCLUDED
