//
// directory_reader_pos.cpp
//
// Bill Seymour, 2024-03-08
//
// This file is in the public domain.
//
// This is the POSIX implementation of the class
// declared in directory_reader.hpp.
//

#include "directory_reader.hpp"

#include <stdexcept>
#include <utility> // move
#include <cstring> // memcpy, memset
#include <cassert>

#include <unistd.h>

#include <sys/param.h>
#ifdef MAXPATHLEN
  #define PATHBUFLEN MAXPATHLEN
#else
  #include <limits.h>
  #ifdef PATH_MAX
    #define PATHBUFLEN PATH_MAX
  #else
    #define PATHBUFLEN 1023
  #endif
#endif

namespace dirrdr {

directory_reader::info_type::info_type()
  : prev_dir(), this_dir(), handle(nullptr), dirp(nullptr)
{
    std::memset(&databuf, '\0', sizeof databuf);
}

directory_reader::info_type::~info_type()
{
    if (handle != nullptr)
    {
        closedir(handle);
        handle = nullptr;
    }
    dirp = nullptr;
}

directory_reader::info_type::info_type(info_type&& rhs) noexcept
  : prev_dir(std::move(rhs.prev_dir)), this_dir(std::move(rhs.this_dir)),
    handle(rhs.handle), dirp(rhs.dirp)
{
    std::memcpy(&databuf, &rhs.databuf, sizeof databuf);

    rhs.handle = nullptr;
    rhs.dirp = nullptr;
    std::memset(&rhs.databuf, '\0', sizeof rhs.databuf);
}

directory_reader::info_type&
  directory_reader::info_type::operator=(info_type&& rhs) noexcept
{
    prev_dir = std::move(rhs.prev_dir);
    this_dir = std::move(rhs.this_dir);
    handle = rhs.handle;
    dirp = rhs.dirp;

    std::memcpy(&databuf, &rhs.databuf, sizeof databuf);

    rhs.handle = nullptr;
    rhs.dirp = nullptr;
    std::memset(&rhs.databuf, '\0', sizeof rhs.databuf);

    return *this;
}

directory_reader& directory_reader::open(const char* new_dir)
{
    info_type new_info;

    if (!data.empty())
    {
        new_info.prev_dir.assign(data.back().this_dir);
    }
    // Else this is the 1st directory that we're opening,
    // so there is no previous one.

    //
    // Now figure out where we're moving to.
    //
    if (new_dir == nullptr || *new_dir == '\0')
    {
        // Use the CWD.
        char buf[PATHBUFLEN + 1];
        char const * const cwd = getcwd(buf, sizeof buf);
        assert(cwd != nullptr);
        new_info.this_dir.assign(cwd);
    }
    else
    {
        // The new directory is specified.
        if (*new_dir == '/')
        {
            // Absolute path.
            new_info.this_dir.assign(new_dir);
        }
        else
        {
            // Relative to the current directory.
            new_info.this_dir.assign(new_info.prev_dir);
            new_info.this_dir.append(1, '/');
            new_info.this_dir.append(new_dir);
        }
    }

    //
    // Now we know which directory to open.
    //
    if ((new_info.handle = opendir(new_info.this_dir.c_str())) == nullptr)
    {
        std::string msg("Can't open directory \"");
        msg.append(new_info.this_dir);
        msg.append(1, '"');
        throw std::invalid_argument(msg);
    }

    if (!new_info.do_read())
    {
        // End of directory right from the get-go (not even . or ..)
        std::string msg("Directory \"");
        msg.append(new_info.this_dir);
        msg.append("\" is totally empty");
        throw std::runtime_error(msg);
    }

    // Success.
    data.push_back(std::move(new_info));
    return *this;
}

directory_reader& directory_reader::open(const std::string& new_dir)
{
    return open(new_dir.c_str());
}

bool directory_reader::info_type::do_read()
{
    struct dirent* dptr = readdir(handle);
    if (dptr != nullptr)
    {
        std::string full_path(this_dir);
        full_path.append(1, '/');
        full_path.append(dptr->d_name);
        if (lstat(full_path.c_str(), &databuf) != 0)
        {
            std::string msg("Can't stat \"");
            msg.append(full_path);
            msg.append(1, '"');
            throw std::runtime_error(msg);
        }
        dirp = dptr;
        return true;
    }
    return false; // end of directory
}

//
// Everything else requires that we have already opened some directory.
//

bool directory_reader::read_next()
{
    assert(!data.empty());
    return data.back().do_read();
}

bool directory_reader::rewind()
{
    assert(!data.empty());
    info_type& info = data.back();
    rewinddir(info.handle);
    return info.do_read();
}

void directory_reader::close()
{
    assert(!data.empty());
    info_type& info = data.back();
    closedir(info.handle);
    info.handle = nullptr;
    data.pop_back();
}

const std::string& directory_reader::directory() const noexcept
{
    assert(!data.empty());
    return data.back().this_dir;
}

const char* directory_reader::entry_name() const noexcept
{
    assert(!data.empty());
    return data.back().dirp->d_name;
}

unsigned long long directory_reader::file_size() const noexcept
{
    assert(!data.empty());
    return static_cast<unsigned long long>(data.back().databuf.st_size);
}

bool directory_reader::is_regular_file() const noexcept
{
    assert(!data.empty());
    return S_ISREG(data.back().databuf.st_mode);
}

bool directory_reader::is_directory() const noexcept
{
    assert(!data.empty());
    return S_ISDIR(data.back().databuf.st_mode);
}

const directory_reader::buftype& directory_reader::more_data() const noexcept
{
    assert(!data.empty());
    return data.back().databuf;
}

} // namespace dirrdr

// End of directory_reader_pos.cpp
