//
// directory_reader_win.cpp
//
// Bill Seymour, 2024-03-08
//
// This file is in the public domain.
//
// This is the implementation for Microsoft Windows
// of the class declared in directory_reader.hpp.
//

#include "directory_reader.hpp"

#include <stdexcept>
#include <utility> // move
#include <cstring> // memset, memcpy
#include <cassert>

#include <direct.h> // _getcwd, _chdir
// The rest of the vendor lock-in stuff is in io.h
// which directory_reader.hpp has already #included.

namespace dirrdr {

directory_reader::info_type::info_type()
  : prev_dir(), this_dir(), handle(-1L)
{
    std::memset(&databuf, '\0', sizeof databuf);
}

directory_reader::info_type::~info_type()
{
    if (handle != -1L)
    {
        _findclose(handle);
        handle = -1L;
    }
}

directory_reader::info_type::info_type(info_type&& rhs) noexcept
  : prev_dir(std::move(rhs.prev_dir)), this_dir(std::move(rhs.this_dir)),
    handle(rhs.handle)
{
    std::memcpy(&databuf, &rhs.databuf, sizeof databuf);

    rhs.handle = -1L;
    std::memset(&rhs.databuf, '\0', sizeof rhs.databuf);
}

directory_reader::info_type&
  directory_reader::info_type::operator=(info_type&& rhs) noexcept
{
    prev_dir = std::move(rhs.prev_dir);
    this_dir = std::move(rhs.this_dir);
    handle = rhs.handle;

    std::memcpy(&databuf, &rhs.databuf, sizeof databuf);

    rhs.handle = -1L;
    std::memset(&rhs.databuf, '\0', sizeof rhs.databuf);

    return *this;
}

directory_reader& directory_reader::open(const char* new_dir)
{
    //
    // The CWD will become the "previous directory" and new_dir,
    // which is what we're opening, will become "this directory".
    //
    info_type new_info;

    //
    // Where we are now.
    //
    if (!data.empty())
    {
        // We already know the CWD.
        new_info.prev_dir.assign(data.back().this_dir);
    }
    else
    {
        // This is the first directory in the stack,
        // so we need to discover the CWD.
        char buf[2048];
        char const * const cwd = _getcwd(buf, sizeof buf);
        assert(cwd != nullptr);
        new_info.prev_dir.assign(cwd);
    }

    //
    // Now figure out where we're moving to.
    //
    if (new_dir == nullptr || *new_dir == '\0' || new_info.prev_dir == new_dir)
    {
        // Stay in the same directory on the same drive.
        new_info.this_dir.assign(new_info.prev_dir);
    }
    else
    {
        // We have a new directory and, maybe, a new drive.
        if (*new_dir == '\\')
        {
            // The root directory on the same drive.
            new_info.this_dir.assign(1, new_info.prev_dir[0]);
            new_info.this_dir.append(1, ':');
            new_info.this_dir.append(new_dir);
        }
        else if (new_dir[1] != ':' || *new_dir == new_info.prev_dir[0])
        {
            // Relative to the current directory on the same drive.
            new_info.this_dir.assign(new_info.prev_dir);
            new_info.this_dir.append(1, '\\');
            new_info.this_dir.append(new_dir);
        }
        else
        {
            // Different drive implies whole new path.
            new_info.this_dir.assign(new_dir);
        }

        // Now change to the new directory.
        if (_chdir(new_info.this_dir.c_str()) != 0)
        {
            std::string msg("No such directory as \"");
            msg.append(new_info.this_dir);
            msg.append(1, '"');
            throw std::invalid_argument(msg);
        }
    }

    //
    // We're where we need to be, so read the first directory entry.
    //
    if ((new_info.handle = _findfirsti64("*.*", &new_info.databuf)) == -1L)
    {
        std::string msg("Can't read the first entry in \"");
        msg.append(new_info.this_dir);
        msg.append(1, '"');
        throw std::runtime_error(msg);
    }

    // Success.
    data.push_back(std::move(new_info));
    return *this;
}

directory_reader& directory_reader::open(const std::string& new_dir)
{
    return open(new_dir.c_str());
}

//
// Everything else requires that we have already opened some directory.
//

bool directory_reader::read_next() noexcept
{
    assert(!data.empty());
    info_type& info = data.back();
    return _findnexti64(info.handle, &info.databuf) == 0;
}

bool directory_reader::rewind() noexcept
{
    assert(!data.empty());

    //
    // Close and reopen.
    //
    info_type& info = data.back();
    _findclose(info.handle);
    info.handle = _findfirsti64("*.*", &info.databuf);

    //
    // It must have worked previously.
    //
    assert(info.handle != -1L);
    return true;
}

void directory_reader::close()
{
    assert(!data.empty());
    info_type& info = data.back();
    _findclose(info.handle);
    info.handle = -1L;
    int oops = _chdir(info.prev_dir.c_str());
    assert(oops == 0); // It must have worked previously.
    data.pop_back();
}

const std::string& directory_reader::directory() const noexcept
{
    assert(!data.empty());
    return data.back().this_dir;
}

const char* directory_reader::entry_name() const noexcept
{
    assert(!data.empty());
    return data.back().databuf.name;
}

unsigned long long directory_reader::file_size() const noexcept
{
    assert(!data.empty());
    return static_cast<unsigned long long>(data.back().databuf.size);
}

bool directory_reader::is_regular_file() const noexcept
{
    assert(!data.empty());
    return (data.back().databuf.attrib & ~_A_RDONLY) == 0;
}

bool directory_reader::is_directory() const noexcept
{
    assert(!data.empty());
    return data.back().databuf.attrib & _A_SUBDIR;
}

const directory_reader::buftype& directory_reader::more_data() const noexcept
{
    assert(!data.empty());
    return data.back().databuf;
}

} // namespace dirrdr

// End of directory_reader_win.cpp
